"""
Common wrapper classes
"""

from systematic.classes import SortableContainer
from seine.address import IPv4Address, IPv6Address, \
                          EthernetMACAddress, FirewireLocalLinkAddress
from systematic.log import Logger, LoggerError

class NetworkError(Exception):
    pass


class NetworkAddress(SortableContainer):
    compare_fields = [ 'address' ]
    def __init__(self, family, address, options=[]):
        self.family = family
        self.address = address
        self.broadcast = None
        self.flags = {}
        self.parse_options(options)

    def parse_options(self, options):
        return

    def __repr__(self):
        return '%s' % self.address


class NetworkInterface(SortableContainer):
    compare_fields = [ 'name' ]

    def __init__(self, name):
        self.name = name
        self.media = None
        self.media_status = None
        self.media_config =  None
        self.media_state = None
        self.mtu = None
        self.ipv4_addresses = []
        self.ipv6_addresses = []
        self.flags = []
        self.options = {}
        self.nd6_options = {}
        self.configuration = {}
        self.vlan_id = None
        self.vlan_device = None
        self.vlan_options = {}

    def __repr__(self):
        return self.name

    def sort_addresses(self):
        self.ipv4_addresses.sort()
        self.ipv6_addresses.sort()

class NetworkInterfaceList(list):
    def update(self):
        return

    def sort(self):
        list.sort(self)
        for interface in self:
            interface.sort_addresses()


class Route(SortableContainer):
    compare_fields = [ 'address' ]

    def __init__(self, table, address, target, flags=None, refs=0, use=0, interface=None, expires=None):
        self.table = table
        self.address = address
        self.target = target
        self.flags = flags
        self.refs = refs
        self.use = use
        self.interface = interface
        self.expires = expires


class RoutingTable(object):
    def __init__(self):
        self.ipv4 = []
        self.ipv6 = []


class ARPTableEntry(SortableContainer):
    compare_fields = [ 'address' ]

    def __init__(self, address, ethernet, interface, expires=None, ifscope=None):
        self.address = IPv4Address(address)
        try:
            self.ethernet = EthernetMACAddress(ethernet)
        except ValueError:
            self.ethernet = None
        self.interface = interface
        self.expires = expires
        self.ifscope = ifscope.strip('[]')

    def __repr__(self):
        return '%s %s' % (self.address, self.ethernet)


class ARPTable(list):
    pass




#!/usr/bin/env python
"""
Parser for gTLD first level whois data formats
"""

from datetime import datetime

from seine.whois.formats import WhoisDataParser, WhoisError

FIELD_MAP = {
    'Domain Name':      'domainname',
    'Registrar':        'registrar',
    'Whois Server':     'whois_servers',
    'Referral URL':     'referral_url',
    'Name Server':      'nameservers',
    'Status':           'status',
    'Updated Date':     'updated',
    'Creation Date':    'created',
    'Expiration Date':  'expires',
    'Sponsoring Registrar IANA ID': 'sponsor',
}

DNS_FIELDS = ( 'nameservers', 'whois_servers', 'domainname', )
DATE_FIELDS = ( 'updated', 'created', 'expires', )
DATE_PARSER = lambda value: datetime.strptime(value, '%d-%b-%Y').date()

class gtld(WhoisDataParser):
    tlds = ( 'com', 'edu', 'gov', 'mil', 'net', 'org', 'arpa', )

    def parse(self, domain, data):
        data = WhoisDataParser.parse(self, domain, data)

        for l in data:
            if not l.startswith(' ') or l.strip()=='':
                continue

            try:
                k, v = [x.strip() for x in l.split(':', 1)]
            except ValueError:
                raise WhoisError('Error parsing line %s' % l)

            try:
                k = FIELD_MAP[k]
            except KeyError:
                raise WhoisError('Unknown field on line %s' % l)

            if k in DATE_FIELDS:
                v = DATE_PARSER(v)

            elif k in DNS_FIELDS:
                v = v.lower()

            self.set(k, v)



#!/usr/bin/env python
"""
Modified Pinger class based on pyip package ping.py. Modified to give exact
times as floats, to support intervals and source address changing.

Author: Ilkka Tuohela, hile@iki.fi

Original ping.py header and license:

pyip is a Python package offering assembling/disassembling of raw ip packet
including ip, udp, and icmp. Also it includes 2 utilities based on raw ip,
traceroute and ping.

pyip is released under PYTHON SOFTWARE FOUNDATION LICENSE VERSION 2, and is
a project inspired by 'ping' written by Jeremy Hylton.

Author: Kenneth Jiang, kenneth.jiang@gmail.com

"""

import os
import sys
import time
import string
import decimal
import icmp
import ip
import socket
import select

MAX_PACKETS = 1000
PAYLOAD = 'seine pinger'

class PingError(Exception):
    pass

class PingSocket(object):
    """
    Wrapper for socket to use with ICMP PING packets.

    If source address is given, the requests are sent using given
    address. The address must be configured on a interface.
    """
    def __init__(self, target, source=None):
        self.target = (socket.gethostbyname(target), 0)
        self.socket = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_ICMP)
        self.socket.setblocking(1)

        if source is not None:
            try:
                self.socket.bind((source, socket.IPPROTO_ICMP))
            except socket.error, (ecode, emsg):
                raise PingError('Error binding to source address %s: %s' % (source, emsg))

    def sendto(self, packet):
        try:
            self.socket.sendto(packet, self.target)
        except socket.error, (ecode, emsg):
            raise PingError(emsg)

    def recvfrom(self, maxbytes):
        return self.socket.recvfrom(maxbytes)

class Pinger(object):
    """
    Pinger instance.

    If source address is not given, default chosen by OS is used.

    Timeout is given as milliseconds, default value is 1000.
    Packet interval is given as milliseconds, default value is 1000.
    """

    def __init__(self, target, source=None, timeout=1000, interval=1000):
        self.socket = PingSocket(target, source)
        self.timeout = int(timeout)
        self.interval = int(interval)
        self.pid = os.getpid()
        self.__reset_counters()

    def __reset_counters(self):
        """
        Internal function to reset result counters between ping() calls
        """
        self.last = 0
        self.sent = 0
        self.times = {}
        self.deltas = []
        self.timed_out = []

    def __send_packet(self):
        """
        Internal function to send icmp ECHO packet to target address.
        """
        buf = icmp.assemble(icmp.Echo(id=self.pid, seq=self.sent, data=PAYLOAD))
        self.times[self.sent] = time.time()
        self.sent += 1
        self.socket.sendto(buf)

    def __recv_packet(self, pkt, when, timeout):
        """
        Internal function to process a received response packet
        """
        try:
            sent = self.times[pkt.get_seq()]
        except KeyError:
            print 'Invalid packet sequence number: %s' % pkt.get_seq()
            return

        if when-sent <= timeout:
            self.deltas.append(float(when-sent)*1000)
        else:
            print '%s timed out' % sent
            self.timed_out.append(sent)

    def ping(self, packets=1):
        """
        Send ping messages, as many as given with packets parameter.

        Returns a dictionary of results, may raise PingError.
        """

        try:
            packets = int(packets)
            if packets < 0 or packets > MAX_PACKETS:
                raise ValueError
        except ValueError:
            raise PingError('Invalid number of packets: %s' % packets)

        interval = float(self.interval)/1000
        timeout = float(self.timeout)/1000
        last_sent = 0
        while 1:
            now = time.time()
            if self.sent < packets and (last_sent+interval)<now:
                self.__send_packet()
                last_sent = now

            if len(self.times) == packets:
                if filter(lambda t: t+timeout>now, self.times.values())==[]:
                    break

            (rd, wt, er) = select.select([self.socket.socket], [], [], timeout)
            if not rd:
                continue
            arrival = time.time()
            try:
                (pkt, who) = self.socket.recvfrom(4096)
            except socket.error:
                continue

            reply_address = ip.disassemble(pkt)
            try:
                reply = icmp.disassemble(reply_address.data)
            except ValueError:
                print 'Invalid ICMP reply packet received'
                continue

            if reply.get_id() != self.pid:
                print 'PID in response does not match'
                continue
            self.__recv_packet(reply, arrival, timeout)

            if self.sent < packets:
                continue

            if len(self.deltas)+len(self.timed_out) == packets:
                break

        received = len(self.deltas)
        loss = (float(packets-received)/float(packets))*100

        if len(self.deltas) == 0:
            summary = {
                'min': None,
                'max': None,
                'average': None,
                'sent': packets,
                'received': len(self.deltas) + len(self.timed_out),
                'packetloss': loss
            }

        else:
            summary = {
                'min': min(self.deltas),
                'max': max(self.deltas),
                'average': reduce(lambda x, y: x+y, self.deltas) / len(self.deltas),
                'sent': packets,
                'received': received,
                'packetloss': loss
            }

        self.__reset_counters()
        return summary



#!/usr/bin/env python
"""
Cache for whois server lists.
"""

import sys
import os
import time
import logging
import random

import errno
import socket
import select

import ConfigParser

from seine.address import IPv4Address, IPv6Address
from seine.dns.tld import TLDCache, DNSCacheError
from seine.whois import WhoisError

CACHE_FILES = (
    '/var/cache/whois/servers.cache',
    '/tmp/whois-%s/servers.cache' % os.geteuid(),
)
SEARCH_DOMAIN = 'whois-servers.net'

WHOIS_PORT = 43
WHOIS_SERVER_TIMEOUT = 15
WHOIS_BUFFER_SIZE = 1024

TLD_REQUIRES_EQUALS = ['com']

logger = logging.getLogger(__file__)


class WhoisServer(object):
    def __init__(self, tld, address, timeout=WHOIS_SERVER_TIMEOUT):
        self.tld = tld
        self.address = address
        self.timeout = float(timeout)
        self.socket = None

    def __del__(self):
        self.close()

    def close(self):
        if self.socket is not None:
            self.socket.close()
        self.socket = None

    def connect(self):
        raise NotImplementedError('Must be implemented in child class')

    def send_query(self, domain):
        raise NotImplementedError('Must be implemented in child class')

    def get_response(self, buffer_size=WHOIS_BUFFER_SIZE):
        raise NotImplementedError('Must be implemented in child class')


class IPv4WhoisServer(WhoisServer):
    def connect(self):
        if self.socket is not None:
            raise WhoisError('Connection already open to %s' % self.address)

        self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        while True:
            try:
                self.socket.connect((self.address, WHOIS_PORT))
            except socket.error, (ecode, emsg):
                if ecode in [errno.EINPROGRESS, errno.EALREADY]:
                    continue
                else:
                    raise WhoisError('Error connecting to %s: %s' % (self.address, emsg) )
            break

        ret = select.select([self.socket], [self.socket], [], 5)
        if len(ret[1]) == 0 and len(ret[0]) == 0:
            self.close()
            raise WhoisError('Timeout waiting for data from %s' % self.address)

    def send_query(self, domain):
        if self.socket is None:
            raise WhoisError('Connection not open to %s' % self.address)

        self.socket.setblocking(1)
        self.socket.settimeout(self.timeout)

        if self.tld in TLD_REQUIRES_EQUALS:
            data = '=%s\r\n' % domain
        else:
            data = '%s\r\n' % domain

        self.socket.send(data)

    def get_response(self, buffer_size=WHOIS_BUFFER_SIZE):
        if self.socket is None:
            raise WhoisError('Connection not open to %s' % self.address)

        out = ''
        while True:
            try:
                data = self.socket.recv(WHOIS_BUFFER_SIZE)
            except socket.timeout:
                self.close()
                raise WhoisError('Request timeout to %s' % self.address)
            except socket.error, (ecode, emsg):
                if ecode in [errno.EINPROGRESS, errno.EALREADY]:
                    continue
                else:
                    self.close()
                    raise WhoisError('Error reading data from %s: %s' % (self.address, emsg))

            # End of data
            if not data:
                break

            out += data

        self.close()

        return out


class IPv6WhoisServer(WhoisServer):
    def connect(self):
        raise NotImplementedError('IPv6 whois queries not yet implemented')

    def send(self, s, data):
        raise NotImplementedError('IPv6 whois queries not yet implemented')

    def get_response(self, s, buffer_size=WHOIS_BUFFER_SIZE):
        raise NotImplementedError('IPv6 whois queries not yet implemented')


class TLDWhoisServerList(object):
    def __init__(self, tld, ipv4_addresses=[], ipv6_addresses=[]):
        self.tld = tld
        self.ipv4_addresses = [IPv4Address(address) for address in ipv4_addresses]
        self.ipv6_addresses = [IPv6Address(address) for address in ipv6_addresses]

    def __repr__(self):
        return 'TLD .%s whois servers: %s %s' % (
            self.tld,
            ','.join(x.ipaddress for x in self.ipv4_addresses),
            ','.join(x.address for x in self.ipv6_addresses),
        )

    def query(self, name, ipv4=True, ipv6=False):
        if ipv4:
            if not self.ipv4_addresses:
                raise WhoisError('No IPv4 addresses for %s whois servers' % self.tld)
            server = IPv4WhoisServer(self.tld, random.choice(self.ipv4_addresses).ipaddress)

        elif ipv6:
            if not self.ipv6_addresses:
                raise WhoisError('No IPv6 addresses for %s whois servers' % self.tld)
            server = IPv6WhoisServer(random.choice(self.ipv6_addresses).address)

        tld = name.rstrip('.').split('.')[-1]
        if tld != self.tld:
            raise WhoisError('TLD %s does not match server: %s' % (tld, self.tld))

        server.connect()
        server.send_query(name)
        return server.get_response()

class WhoisServerCache(dict):
    def __init__(self, cache_path=None, tld_cache_path=None):
        self.cache_path = None

        if cache_path is not None:
            self.cache_path = os.path.expandvars(os.path.expanduser(cache_path))

        else:
            for f in CACHE_FILES:
                if os.path.isfile(f) and os.access(f, os.W_OK):
                    self.cache_path = f
                    break

                fdir = os.path.dirname(f)
                if not os.path.isdir(fdir):
                    try:
                        os.makedirs(os.path.dirname(f))
                    except IOError, (ecode, emsg):
                        continue
                    except OSError, (ecode, emsg):
                        continue

                if not os.path.isfile(f):
                    try:
                        open(f, 'w').write('\n')
                        os.unlink(f)
                        self.cache_path = f
                        break
                    except IOError, (ecode, emsg):
                        continue
                    except OSError, (ecode, emsg):
                        continue

        if self.cache_path is None:
            raise WhoisError('ERROR: No whois cache path defined')

        try:
            self.tlds = TLDCache(tld_cache_path)
            if not self.tlds.is_downloaded:
                self.tlds.download()
        except DNSCacheError, emsg:
            raise WhoisEror(emsg)

        if os.path.isfile(self.cache_path):
            self.load()

    def load(self):
        if not os.path.isfile(self.cache_path):
            raise WhoisError('No such file: %s' % self.cache_path)

        try:
            cache = ConfigParser.ConfigParser()
            cache.read(self.cache_path)
        except OSError, (ecode, emsg):
            raise WhoisError(emsg)
        except IOError, (ecode, emsg):
            raise WhoisError(emsg)

        for tld in cache.sections():
            try:
                ipv4_addresses = []
                ipv6_addresses = []
                if cache.has_option(tld, 'ipv4_addresses') and cache.get(tld, 'ipv4_addresses'):
                    ipv4_addresses.extend(cache.get(tld, 'ipv4_addresses').split(','))
                if cache.has_option(tld, 'ipv6_addresses') and cache.get(tld, 'ipv6_addresses'):
                    ipv6_addresses.extend(cache.get(tld, 'ipv6_addresses').split(','))
                entry = TLDWhoisServerList(tld, ipv4_addresses, ipv6_addresses)
                self[tld] = entry
            except ConfigParser.NoOptionError, emsg:
                logger.debug('Invalid cache entry for %s' % tld)

    def save(self):
        if len(self) == 0:
            return

        cache_dir = os.path.dirname(self.cache_path)

        if not os.path.isdir(cache_dir):
            try:
                os.makedirs(cache_dir)
            except IOError, (ecode, emsg):
                raise WhoisError('Error creating cache directory %s: %s' % (cache_dir, emsg))
            except OSError, (ecode, emsg):
                raise WhoisError('Error creating cache directory %s: %s' % (cache_dir, emsg))

        try:
            cache = ConfigParser.ConfigParser()
            for entry in self.values():
                section = cache.add_section(entry.tld)
                cache.set(entry.tld, 'ipv4_addresses', ','.join(x.ipaddress for x in entry.ipv4_addresses))
                cache.set(entry.tld, 'ipv6_addresses', ','.join(x.address for x in entry.ipv6_addresses))
            cache.write(open(self.cache_path, 'w'))

        except OSError, (ecode, emsg):
            raise WhoisError('Error updating cache %s: %s' % (self.cache_path, emsg))
        except IOError, (ecode, emsg):
            raise WhoisError('Error updating cache %s: %s' % (self.cache_path, emsg))

    def get(self, tld, timeout=WHOIS_SERVER_TIMEOUT):
        try:
            timeout = int(timeout)
        except ValueError:
            raise WhoisError('Invalid timeout value: %s' % timeout)

        if tld not in self.tlds:
            raise WhoisError('Invalid TLD name: %s' % tld)

        if tld not in self:
            details = {}
            name = '.'.join([tld, SEARCH_DOMAIN])
            try:
                socket.setdefaulttimeout(timeout)
                name, aliases, addresses = socket.gethostbyname_ex(name)
                entry = TLDWhoisServerList(tld, addresses)
                self[tld] = entry

            except socket.timeout:
                raise WhoisError('Timeout resolving whois server %s' % name)
            except socket.gaierror, (ecode, emsg):
                raise WhoisError(emsg)
            except ValueError, emsg:
                raise WhoisError(emsg)

        return self[tld]

    def query(self, name):
        try:
            tld = name.rstrip('.').split('.')[-1]
        except ValueError:
            raise WhoisError('Error parsing TLD from %s' % name)

        tldservers = self.get(tld)
        return tldservers.query(name)



#!/usr/bin/env python
"""
Parser classes for various whois output formats
"""

import os

from seine.whois import WhoisError
from seine.whois.formats.ficora import ficora
from seine.whois.formats.gtld import gtld
from seine.whois.formats.nominet import nominet
from seine.whois.formats.saudinic import saudinic

WHOIS_PARSERS = (
    ficora, gtld, nominet, saudinic
)


class WhoisData(dict):
    def __init__(self, domain):
        self.parsers = WHOIS_PARSERS
        self.domain = domain

    def __repr__(self):
        return 'WHOIS for %s' % self.domain

    @property
    def name(self):
        try:
            return self['domainname']
        except KeyError:
            return self.domain

    @property
    def status(self):
        try:
            return self['status']
        except KeyError:
            return []

    @property
    def nameservers(self):
        try:
            return sorted(x.lower() for x in self['nameservers'])
        except KeyError:
            return []

    @property
    def created(self):
        try:
            return self['created']
        except KeyError:
            return None

    @property
    def expires(self):
        try:
            return self['expires']
        except KeyError:
            return None

    @property
    def modified(self):
        try:
            return self['modified']
        except KeyError:
            return None

    def query(self, data):
        data_parsers = [x for x in self.parsers if x().matches_domain(self.domain)]
        if len(data_parsers) > 1:
            raise WhoisError('BUG: more than one parser for domain %s: %s' % (
                self.domain,
                [x.name for x in data_parsers]
            ))

        if not data_parsers:
            raise WhoisError('No whois data parser found for domain %s' % self.domain)

        formatter = data_parsers[0]()
        formatter.parse(self.domain, data)
        self.update(formatter.items())






import os
import re
from subprocess import check_output, CalledProcessError

from seine.address import IPv4Address, IPv6Address, \
                          EthernetMACAddress, FirewireLocalLinkAddress
from seine.classes import NetworkInterface, NetworkAddress, NetworkInterfaceList, \
                          Route, RoutingTable, ARPTable, ARPTableEntry, NetworkError

RE_INTERFACE = re.compile('^\d+: (?P<name>[^:]+): <(?P<flags>[^>]+)> mtu (?P<mtu>\d+) (?P<options>.*)$')
ARP_ENTRY_FORMATS = (
    re.compile('^\? \((?P<address>[0-9.]+)\) at (?P<ethernet>[^\s]+) (?P<ifscope>.*) on (?P<interface>[^\s]+)$'),
)

NETSTAT_HEADER_MATCHES = {
    re.compile('^Kernel IP.* routing table$'),
    re.compile('^Destination\s+.*$'),
}


def parse_ipv4_route_target(value, netmask=None):
    if value == 'default':
        return IPv4Address('0.0.0.0/0')

    elif value[:5] == 'link#':
        return value

    else:
        try:
            return EthernetMACAddress(value)
        except ValueError:
            pass

    return IPv4Address(value, netmask)


def parse_ipv6_route_target(value):
    if value == 'default':
        return IPv6Address('0::/0')

    elif value[:5] == 'link#':
        return value

    else:
        try:
            return EthernetMACAddress(value)
        except ValueError:
            pass

    return IPv6Address(value)


class LinuxNetworkAddress(NetworkAddress):
    def __init__(self, family, address, options=[]):
        NetworkAddress.__init__(self, family, address, options)

    def parse_options(self, options):
        if self.family == 'inet':
            option = None

            while options:
                if options[0] == 'scope':
                    self.flags['scope'] = options[1]
                    options = options[3:]

                elif options[0] == 'brd':
                    self.broadcast = IPv4Address('%s/32' % options[1])
                    options = options[2:]

                else:
                    options = options[1:]

        if self.family == 'inet6':
            option = None

            while options:
                if options[0] == 'scope':
                    if options[1] in ('host', 'link'):
                        self.flags['scope'] = options[1]
                        options = ' '.join(options[2:])
                    elif options[1] == 'global':
                        self.flags['scope'] = ' '.join(options[1:3])
                        options = options[3:]
                    else:
                        options = options[1:]
                else:
                    options = options[1:]


class LinuxNetworkInterface(NetworkInterface):
    def __init__(self, name):
        NetworkInterface.__init__(self, name)
        self.__latest_parsed_address = None

        if self.name.count('@') == 1:
            self.parse_vlan_options()

    def parse_vlan_options(self):
        self.name, self.vlan_device = self.name.split('@', 1)
        try:
            output = check_output(['/sbin/ip', '-d', 'link', 'show', self.name])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting interfaces: %s' % emsg)

        for l in [l for l in output.split('\n')[1:] if l.strip() != '']:
            fields = l.split()
            if fields[0] != 'vlan':
                continue

            while fields:
                if fields[0] == 'id':
                    self.vlan_id = fields[1]
                    fields = fields[2:]
                if fields[0] in ( 'protocol', ):
                    self.vlan_options[fields[0]] = fields[1]
                    fields = fields[2:]
                elif fields[0].startswith('<'):
                    self.vlan_options['flags'] = fields[0].strip('<>').split(',')
                    fields = fields[1:]
                else:
                    fields = fields[1:]

    def parse_options(self, fields):
        fields = fields.split()

        while fields:
            if fields[0] == 'state':
                if fields[1] == 'UNKNOWN':
                    self.media_state = None
                else:
                    self.media_state = fields[1]
                self.options['state'] = fields[1]
                fields = fields[2:]

            elif fields[0] == 'qlen':
                self.options['qlen'] = int(fields[1])
                fields = fields[2:]

            else:
                self.options[fields[0]] = True
                fields = fields[1:]

    def parse(self, line):
        fields = line.strip().split()

        if not fields:
            return

        if fields[0] == 'inet':
            address = LinuxNetworkAddress('inet', fields[1], fields[2:])
            self.ipv4_addresses.append(address)
            self.__latest_parsed_address = address

        elif fields[0] == 'inet6':
            address = LinuxNetworkAddress('inet6', fields[1], fields[2:])
            self.ipv6_addresses.append(address)
            self.__latest_parsed_address = address

        elif fields[0] == 'link/ether':
            self.media = EthernetMACAddress(fields[1])

        elif fields[0] == 'link/loopback':
            self.media = EthernetMACAddress(fields[1])

        elif fields[0] == 'valid_lft':
            address = self.__latest_parsed_address
            address.flags['valid_lft'] = fields[1]
            address.flags['preferred_lft'] = fields[3]


class LinuxNetworkInterfaces(NetworkInterfaceList):
    def update(self):
        self.__delslice__(0, len(self))

        try:
            output = check_output(['/sbin/ip', 'addr', 'show'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting interfaces: %s' % emsg)

        interface = None
        for l in [l for l in output.split('\n') if l.strip() != '']:
            m = RE_INTERFACE.match(l)
            if m:
                interface = LinuxNetworkInterface(m.groupdict()['name'])
                interface.mtu = m.groupdict()['mtu']
                interface.flags = m.groupdict()['flags'].split(',')
                interface.parse_options(m.groupdict()['options'])
                self.append(interface)

            else:
                interface.parse(l)


class IPv4RouteEntry(Route):
    def __init__(self, table, line):
        fields = line.split()

        address = parse_ipv4_route_target(fields[0], fields[2])
        target = parse_ipv4_route_target(fields[1])
        flags = fields[4]
        interface = fields[7]

        Route.__init__(self, table, address, target, flags=flags, interface=interface)

    def __str__(self):
        return '%s %s' % (self.address, self.target)


class IPv6RouteEntry(Route):
    def __init__(self, table, line):
        fields = line.split()

        address = parse_ipv6_route_target(fields[0])
        target = parse_ipv6_route_target(fields[1])
        flags = fields[2]
        refs = fields[4]
        use = fields[5]
        interface = fields[6]

        Route.__init__(self, table, address, target, flags=flags, interface=interface, refs=refs, use=use)

    def __str__(self):
        return '%s %s' % (self.address, self.target)


class LinuxRoutes(RoutingTable):
    """Linux style netstat parsing

    Parse linux style netstat -rn outputs

    """
    def update(self):
        self.ipv4 = []
        self.ipv6 = []

        try:
            output = check_output(['/bin/netstat', '-rn', '-4'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting IPv4 routes: %s' % emsg)

        interface = None
        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.strip() == '':
                continue
            header = False
            for m in NETSTAT_HEADER_MATCHES:
                if m.match(l):
                    header = True
                    break
            if header:
                continue

            self.ipv4.append(IPv4RouteEntry(self, l))

        try:
            output = check_output(['/bin/netstat', '-rn', '-6'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting IPv6 routes: %s' % emsg)

        interface = None
        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.strip() == '':
                continue
            header = False
            for m in NETSTAT_HEADER_MATCHES:
                if m.match(l):
                    header = True
                    break
            if header:
                continue

            self.ipv6.append(IPv6RouteEntry(self, l))

        self.sort()

    def sort(self):
        self.ipv4.sort()
        self.ipv6.sort()


class LinuxARP(ARPTable):
    def update(self):
        self.__delslice__(0, len(self))

        try:
            output = check_output(['/usr/sbin/arp', '-an'])
        except CalledProcessError, emsg:
            raise NetworkError('Error getting ARP table: %s' % emsg)

        for l in [l for l in output.split('\n') if l.strip() != '']:
            if l.strip() == '':
                continue

            for fmt in ARP_ENTRY_FORMATS:
                m = fmt.match(l)
                if m:
                    self.append(ARPTableEntry(**m.groupdict()))



"""
Classes to represent Ethernet, IPv4, IPv6 addresses and address ranges
"""

import struct
import string

from systematic.classes import SortableContainer

# Maximum value available with 32 bits
UINT_MAX = 2**32-1
U128_MAX = 2**128-1

ADDRESS_CLASS_DEFAULT = 'normal'

IPV4_ADDRESS_CLASS_MAP = {
    'this_broadcast': {
        'networks': ['0.0.0.0/8 '],
        'description': 'Used for broadcast messages to the current ("this") network as specified by RFC 1700, page 4.',
    },
    'carriernat':           {
        'networks': ['100.64.0.0/10'],
        'description': 'Used for communications between a service provider and its subscribers when using a Carrier-grade NAT, as specified by RFC 6598',
    },

    'loopback':             {
        'networks': ['127.0.0.0/8'],
        'description': 'Used for loopback addresses to the local host, as specified by RFC 990',
    },
    'link-local':           {
        'networks': ['169.254.0.0/16'],
        'description': 'Used for link-local addresses between two hosts on a single link when no IP address is otherwise specified, such as would have normally been retrieved from a DHCP server, as specified by RFC 3927',
    },
    'spar':                 {
        'networks': ['192.0.0.0/24'],
        'description': 'Used for the IANA IPv4 Special Purpose Address Registry as specified by RFC 5736',
    },
    'testnet':              {
        'networks': ['192.0.2.0/24'],
        'description': 'Assigned as "TEST-NET" in RFC 5737 for use solely in documentation and example source code and should not be used publicly',
    },
    '6to4anycast':          {
        'networks': ['192.88.99.0/24'],
        'description': 'Used by 6to4 anycast relays as specified by RFC 3068',
    },
    'test_inter_network':     {
        'networks': ['198.18.0.0/15'],
        'description': 'Used for testing of inter-network communications between two separate subnets as specified in RFC 2544',
    },
    'testnet_2':            {
        'networks': ['198.51.100.0/24'],
        'description': 'Used for testing of inter-network communications between two separate subnets as specified in RFC 2544',
    },
    'testnet_3':            {
        'networks': ['203.0.113.0/24'],
        'description': 'Assigned as "TEST-NET-3" in RFC 5737 for use solely in documentation and example source code and should not be used publicly.',
    },
    'rfc1918':              {
        'networks': ['10.0.0.0/8', '172.16.0.0/12', '192.168.0.0/16'],
        'description': 'Used for local communications within a private network as specified by RFC 1918',
    },
    'multicast':            {
        'networks': ['224.0.0.0/4'],
        'description': 'Reserved for multicast assignments as specified in RFC 5771. 233.252.0.0/24 is assigned as "MCAST-TEST-NET" for use solely in documentation and example source code.',
    },
    'reserved':             {
        'networks': ['240.0.0.0/4'],
        'description': 'Reserved for future use, as specified by RFC 6890',
    },
    'limited_broadcast':    {
        'networks': ['255.255.255.255/32'],
        'description': 'Reserved for the "limited broadcast" destination address, as specified by RFC 6890',
    },
}
IPV6_ADDRESS_CLASS_MAP = {
    'undefined':                {
        'networks': ['::/128'],
        'description': 'Unspecified address',
    },
    'loopback':                 {
        'networks': ['::1/128'],
        'description': 'loopback address to the local host',
    },
    'local_ipv4_translation':   {
        'networks': ['::ffff:0:0/96'],
        'description': 'IPv4 mapped addresses',
    },
    'discard_prefix_rfc6666':   {
        'networks': ['100::/64'],
        'description': 'Discard Prefix RFC 6666',
    },
    'global_ipv4_translation':  {
        'networks': ['64:ff9b::/96'],
        'description': 'IPv4/IPv6 translation RFC 6052',
    },
    'teredo':                   {
        'networks': ['2001::/32'],
        'description': 'Teredo tunneling',
    },
    'orchid_deprecated':        {
        'networks': ['2001:10::/28'],
        'description': 'Deprecated (previously ORCHID)',
    },
    'orchid_v2':                {
        'networks': ['2001:20::/28'],
        'description': 'ORCHIDv2',
    },
    'documentation':            {
        'networks': ['2001:db8::/32'],
        'description': 'Addresses used in documentation',
    },
    '6to4':                     {
        'networks': ['2002::/16'],
        'description': '6to4 tunneling',
    },
    'unique_local':             {
        'networks': ['fc00::/7'],
        'description': 'Unique local address',
    },
    'link_local':               {
        'networks': ['fe80::/10'],
        'description': 'Link-local address',
    },
    'multicast':                {
        'networks': ['ff00::/8'],
        'description': 'Multicast',
    },
}

IPV6_IPV4_TRANSLATION_PREFIXES = (
    '::ffff',
    '64:ff9b',
)

try:
    type(bin)
except NameError:
    def bin(str,pad32bits=True):
        if type(str) not in [int,long]:
            str = long(str)
        t={
            '0':'000','1':'001','2':'010','3':'011',
            '4':'100','5':'101','6':'110','7':'111'
        }
        s=''
        for c in oct(str).rstrip('L')[1:]:
            s+=t[c]
        s = s.lstrip('0')
        if pad32bits and len(s) < 32:
            s = '0'*(32-len(s)) + s
        return s


def isEthernetMACAddress(value):
    try:
        EthernetMACAddress(value)
    except ValueError:
        return False
    return True


class MediaAddressType(SortableContainer):
    """MAC address types base class

    Common base class for MAC address types, like Ethernet, Firewire etc.

    """
    def __init__(self, addresstype, address):
        self.type = addresstype
        self.__parseaddress__(address)

    def __parseaddress__(self, value):
        raise NotImplementedError('__parseaddress__ must be implemented in child class')

    def __repr__(self):
        return self.address

    def __hash__(self):
        return self.value

    def __int__(self):
        return self.value

    def __long__(self):
        return self.value

    def __cmp__(self, other):
        if self.__class__ == other.__class__:
            return cmp(self.value, other.value)

        elif isinstance(other, MediaAddressType):
            raise ValueError('Error comparing %s to %s' % (
                    self.__class__.__name__,
                    other.__class__.__name__,
                ))

        else:
            try:
                other = self.__class__(other)
                return cmp(self.value, other.value)

            except ValueError:
                raise ValueError('Compared value is not valid %s value: %s' % (
                    self.__class__.__name__,
                    other
                ))


class EthernetMACAddress(MediaAddressType):
    def __init__(self, address):
        MediaAddressType.__init__(self, 'ethernet', address)

    def __parseaddress__(self, value):
        try:
            if len(value) == 12:
                parts = [int(x, 16) for x in
                    [value[i:i+2] for i in range(0, len(value), 2)]
                ]

            elif len(value) == 6:
                parts = struct.unpack('BBBBBB', str(value))

            else:
                parts = [int(x, 16) for x in value.split(':', 5)]

            if len(parts) != 6:
                raise ValueError

            for p in parts:
                if p < 0 or p > 255:
                    raise ValueError

        except ValueError:
            raise ValueError('Not a Ethernet MAC address: %s' % value)

        self.address = ':'.join(['%02x'%p for p in parts])
        self.value = sum([parts[-(len(parts)-i)]<<8*(len(parts)-i-1) for i in range(len(parts))])

class FirewireLocalLinkAddress(MediaAddressType):
    def __init__(self, address):
        MediaAddressType.__init__(self, 'firewire', address)

    def __parseaddress__(self, value):
        try:
            if len(value) == 16:
                parts = [int(x, 16) for x in
                    [value[i:i+2] for i in range(0, len(value), 2)]
                ]

            elif len(value) == 8:
                parts = struct.unpack('BBBBBBBB', str(value))

            else:
                parts = [int(x, 16) for x in value.split(':', 7)]

            if len(parts) != 8:
                raise ValueError

            for p in parts:
                if p < 0 or p > 255:
                    raise ValueError

        except ValueError:
            raise ValueError('Not a Firewire local link address: %s' % value)

        self.address = ':'.join(['%02x'%p for p in parts])
        self.value = sum([parts[-(len(parts)-i)]<<8*(len(parts)-i-1) for i in range(len(parts))])


class IPv4Address(object):
    """
    Verify and format IPv4 address given in n.n.n.n/32 format,
    calculate various values for the address.

    Raises ValueError if address is not valid.

    Attributes available:
    ipaddress: x.x.x.x address
    bitmask:   bitmask (0-32)
    netmask:   netmask in x.x.x.x format
    inverted_netmask: netmask in x.x.x.x, inverted (cisco style)
    network:   network address, raises ValueError for /32 addresses
    broadcast: broadcast address, raises ValueError for /32 addresses
    first:     return first host address in network
    last:      return last host address in network

    Internally available:
    address:   IPv4 address as long integer
    mask:      IPv4 netmask as long integer
    """

    def __init__(self, address, netmask=None, oldformat=False):
        """
        Parameters:
        address: dot format address as in inet, or long integer
        netmask: netmask in dot format or long integer
        oldformat: parse 127 as 127.0.0.0 not 0.0.0.127 (as in netstat output)
        """
        self.oldformat = oldformat
        if type(address) != int and len(address) == 4 and not address.translate(None, string.digits+'abcdef'):
            address = '.'.join(str(x) for x in struct.unpack('BBBB', str(address)))
            mask = 32

        elif isinstance(address, basestring) and address[:2] == '0x':
            address = long(address, 16)
            mask = 32

        if type(address) in [int,long]:
            ip = address
            mask = 32

        else:
            try:
                (ip,mask) = address.split('/', 1)
            except ValueError:
                ip = address.strip()
                mask = 32

        if netmask:
            try:
                netmask = self.__parseaddress__(netmask)
                if netmask == UINT_MAX:
                    mask = 32
                else:
                    if bin(UINT_MAX &~ netmask)[2:].count('0')>0:
                        raise ValueError
                    mask = 32 - len(bin(UINT_MAX &~ netmask)) + 2

            except ValueError:
                raise ValueError('Invalid netmask value: %s' % netmask)

        elif self.oldformat:
            if address.count('.') == 2:
                mask = 24
            elif address.count('.') == 1:
                mask = 16
            elif address.count('.') == 0:
                mask = 8
            else:
                mask = 32

        try:
            mask = int(mask)
            if mask not in range(0,33):
                raise ValueError
            self.mask = UINT_MAX ^ (2**(32-mask)-1)
        except ValueError:
            raise ValueError('Invalid netmask: %s' % mask)

        try:
            self.raw_value = self.__parseaddress__(ip)
        except ValueError:
            if isinstance(address,basestring) and address=='default':
                self.raw_value = self.__parseaddress__('0.0.0.0')
                self.mask = 0
            else:
                raise ValueError('Invalid address: %s' % address)

    def __parsenumber__(self, value):
        """
        Parses decimal, octal, hex value from string
        """
        value = str(value)
        if value[:2] == '0x':
            if not value[2:].translate(None, string.digits):
                return int(value, 16)

        elif value[:1] == '0':
            if not value.translate(None, string.digits):
                return int(value, 8)

        else:
            return int(value)

        raise ValueError('Invalid number %s' % value)

    def __parseaddress__(self, value):
        """
        Try to parse an ip-address from various crazy formats defined
        for IP addresses. Of course, sane people would only pass normal
        addresses to us but who knows...
        """
        value = str(value)

        if value.count('.') == 3:
            dotted = []
            parts = value.split('.')
            for i,p in enumerate(parts):
                p = self.__parsenumber__(p)
                if p not in range(0, 256):
                    raise ValueError
                dotted.append(p)

            return reduce(lambda x,y:x+y,[
                (dotted[0]<<24), (dotted[1]<<16), (dotted[2]<<8), (dotted[3]),
            ])

        elif value.count('.') == 2:
            dotted = []
            parts = value.split('.')
            for i,p in enumerate(parts):
                p = self.__parsenumber__(p)
                if not self.oldformat:
                    if i>2 and (p<0 or p>2**8):
                        raise ValueError
                    elif i==2 and (p<0 or p>2**16):
                        raise ValueError
                else:
                    if p<0 or p>2**8:
                        raise ValueError
                dotted.append(p)
            if not self.oldformat:
                return reduce(lambda x,y:x+y,[
                    (dotted[0]<<24), (dotted[1]<<16), (dotted[2])
                ])
            else:
                return reduce(lambda x,y:x+y,[
                    (dotted[0]<<24), (dotted[1]<<16), (dotted[2]<<8)
                ])

        elif value.count('.') == 1:
            dotted = []
            parts = value.split('.')
            for i,p in enumerate(parts):
                p = self.__parsenumber__(p)
                if not self.oldformat:
                    if i==0 and (p<0 or p>2**8):
                        raise ValueError
                    elif i==1 and (p<0 or p>2**24):
                        raise ValueError
                else:
                    if (p<0 or p>2**8):
                        raise ValueError
                dotted.append(p)
            if not self.oldformat:
                return reduce(lambda x,y: x+y, [(dotted[0]<<24), (dotted[1])])
            else:
                return reduce(lambda x,y: x+y, [(dotted[0]<<24), (dotted[1]<<16)])

        elif value.count(' ') == 3:
            # Try 'aa bb cc dd' format hex address conversion. Silly? Yes
            dotted = []
            try:
                dotted = [int(x.strip(), 16) for x in value.split(' ')]
                return reduce(lambda x,y:x+y,[
                    (dotted[0]<<24), (dotted[1]<<16), (dotted[2]<<8), (dotted[3]),
                ])
            except ValueError:
                pass

        else:
            if not self.oldformat:
                return self.__parsenumber__(value)
            else:
                return self.__parsenumber__(value) << 24

        raise ValueError

    def __repr__(self):
        return self.ipaddress

    def __str__(self):
        """
        Returns a CIDR address formatted string for this address
        """
        return self.cidr_address

    def __len__(self):
        """
        Return number of hosts possible in the network, excluding network
        address and broadcast address: NOT reserving a gateway address!
        """
        if self.bitmask > 30:
            return 1
        elif self.bitmask == 30:
            return 2

        first = (self.raw_value & self.mask) + 1
        last  = (self.raw_value & self.mask) + (UINT_MAX &~ self.mask)
        return last-first

    def __hash__(self):
        return long(self.raw_value)

    def __cmp__(self, other):
        if isinstance(other, basestring):
            other = IPv4Address(other)
            return cmp(self.raw_value, other.raw_value)

        elif hasattr(other, 'raw_value'):
            return cmp(self.raw_value, other.raw_value)

        elif isinstance(other, int):
            return cmp(self.raw_value, other)

        else:
            raise ValueError("Can't compare IPv4Address to %s" % type(other))

    def __eq__(self, other):
        return self.__cmp__(other) == 0

    def __ne__(self, other):
        return self.__cmp__(other) != 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __lte__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __gte__(self, other):
        return self.__cmp__(other) >= 0

    def __long__(self):
        """
        Return the integer representation for this IPv4 address
        """
        return self.raw_value

    def __long2address__(self, value):
        """
        Convert a long integer back to n.n.n.n format
        """
        parts = []
        for i in range(0, 4):
            p = str((value &~ (UINT_MAX^2**(32-i*8)-1)) >> (32-(i+1)*8))
            parts.append(p)

        return '.'.join(parts)

    def __addressclass__(self):
        for aclass, details in IPV4_ADDRESS_CLASS_MAP.items():
            for net in details['networks']:
                if IPv4Address(net).addressInNetwork(self.ipaddress):
                    return aclass

        return ADDRESS_CLASS_DEFAULT

    @property
    def address(self):
        return self.__long2address__(self.raw_value)

    @property
    def ipaddress(self):
        return self.__long2address__(self.raw_value)

    @property
    def cidr_address(self):
        if self.bitmask==32:
            return '%s' % self.ipaddress
        else:
            return '%s/%s' % (self.ipaddress, self.bitmask)

    @property
    def bitstring(self):
        return '0x%s' % ''.join(self.hexbytes)

    @property
    def hexbytes(self):
        return map(lambda x:
            '%02x' % int(x),
            self.__long2address__(self.raw_value).split('.')
        )

    @property
    def addressclass(self):
        return self.__addressclass__()

    @property
    def bitmask(self):
        if self.mask == UINT_MAX:
            return 32
        return 32-len(bin(UINT_MAX &~ self.mask))+2

    @property
    def netmask(self):
        return self.__long2address__(self.mask)

    @property
    def inverted_netmask(self):
        return self.__long2address__(UINT_MAX ^ self.mask)

    @property
    def network(self):
        if self.bitmask == 32:
            self.ipaddress
        return self.__long2address__(self.raw_value & self.mask)

    @property
    def broadcast(self):
        if self.bitmask == 32:
            raise ValueError('No broadcast address for /32 address')
        return self.__long2address__(
            (self.raw_value & self.mask) + (UINT_MAX &~ self.mask)
        )

    @property
    def first(self):
        if self.bitmask == 32:
            return self.ipaddress
        if self.bitmask == 31:
            return self.__long2address__(
                self.raw_value & self.mask,
                netmask=self.netmask,
            )
        return IPv4Address(
            (self.raw_value & self.mask) + 1,
            netmask=self.netmask
        )

    @property
    def last(self):
        if self.bitmask == 32:
            return self.ipaddress
        if self.bitmask == 31:
            return self.__long2address__((self.raw_value & self.mask)+1)
        return IPv4Address(
            (self.raw_value & self.mask) + (UINT_MAX &~ self.mask) - 1,
            netmask=self.netmask
        )

    @property
    def next(self):
        address = self.raw_value+1
        if address >= UINT_MAX:
            return None
        return IPv4Address(address, netmask=self.netmask)

    @property
    def prev(self):
        address = self.raw_value-1
        if address < 0:
            return None
        return IPv4Address(address, netmask=self.netmask)

    @property
    def next_network(self):
        address = self.raw_value+2**(32-self.bitmask)
        if address >= UINT_MAX:
            return None
        return IPv4Address('%s/%s' % (address, self.bitmask))

    @property
    def previous_network(self):
        address = self.raw_value-2**(32-self.bitmask)
        if address < 0:
            return None
        return IPv4Address('%s/%s' % (address, self.bitmask))

    @property
    def dns_reverse_ptr(self):
        return '%s.in-addr.arpa.' % '.'.join(reversed(self.ipaddress.split('.')))

    @property
    def dns_reverse_origin(self):
        if self.bitmask >= 24:
            return '%s.in-addr.arpa.' % '.'.join(reversed(self.ipaddress.split('.')[:3]))
        elif self.bitmask >= 16:
            return '%s.in-addr.arpa.' % '.'.join(reversed(self.ipaddress.split('.')[:2]))
        elif self.bitmask >= 8:
            return '%s.in-addr.arpa.' % self.ipaddress.split('.')[0]
        else:
            raise ValueError("Can't create reverse origin for mask %s" % self.bitmask)

    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except TypeError:
            raise KeyError
        except AttributeError, e:
            raise KeyError('No such IPv4Address item: %s' % item)

    def addressInNetwork(self, ip):
        """
        Tests if given IPv4 address is in range of this network,
        including network and broadcast addresses
        """
        if type(ip) != IPv4Address:
            ip = IPv4Address(ip)
        if self.bitmask == 0:
            return True

        if self.bitmask == 32 and ip.raw_value != self.raw_value:
            return False
        else:
            first = self.raw_value & self.mask
            last = (self.raw_value & self.mask) + (UINT_MAX &~ self.mask)
            if ip.raw_value < first or ip.raw_value > last:
                return False

        return True

    def hostInNetwork(self, address):
        """
        Tests if given IPv4 address is in range of this network,
        excluding network and broadcast addresses
        """
        ip = IPv4Address(address)
        if self.bitmask == 0:
            return True
        if self.bitmask == 32 and ip.raw_value != self.raw_value:
            return False

        if self.bitmask == 31:
            first = self.raw_value & self.mask
            if ip.raw_value < first or ip.raw_value > first+1:
                return False
        else:
            first = self.raw_value & self.mask
            last = (self.raw_value & self.mask) + (UINT_MAX &~ self.mask)
            if ip.raw_value <= first or ip.raw_value >= last:
                return False

        return True

    def split(self, bitmask, maxcount=None):
        if self.bitmask >= 30:
            raise ValueError("Can't split network with mask %s" % self.bitmask)

        try:
            bitmask = int(bitmask)
            if bitmask < 1 or bitmask > 30:
                raise ValueError
        except ValueError:
            raise ValueError('Invalid split mask: %s' % bitmask)

        if bitmask <= self.bitmask:
            raise ValueError('Split mask must be larger than network mask %s' % self.bitmask)

        last = self.last

        networks = [IPv4Address('%s/%s' % (self.ipaddress, bitmask))]
        next = IPv4Address('%s/%s' % (self.raw_value+2**(32-bitmask), bitmask))
        while True:
            if maxcount is not None and maxcount < len(networks):
                break
            networks.append(next)

            if next.last.raw_value >= last.raw_value:
                break
            next = IPv4Address('%s/%s' % (next.raw_value + 2**(32-bitmask), bitmask))

        return networks

class IPv4AddressRangeList(list):
    """
    Parses nmap style address range specifications to IPv4AddressRange
    objects. Supported example lines:
        10.0.0.1-254
        10.0.1,2,3.1-254
        10,11.0.1,2,3-5.1-254
    """

    def __init__(self, value):
        try:
            (parts) = value.split('.', 3)
        except ValueError:
            raise ValueError('Unsupported value: %s' % value)

        part_lists = []
        for i, field in enumerate(parts[:-1]):
            part_lists.append([])
            if field.count(',')>0:
                values = field.split(',')
            else:
                values = [field]

            for v in values:
                if v.count('-')==1:
                    start, end = [int(x) for x in v.split('-')]
                    if (start>end): raise ValueError
                    for j in range(start, end+1):
                        part_lists[i].append(j)
                else:
                    part_lists[i].append(v)

        part_lists.append(parts[-1].split(','))
        for p1 in [str(x) for x in part_lists[0]]:
            for p2 in [str(x) for x in part_lists[1]]:
                for p3 in [str(x) for x in part_lists[2]]:
                    for p4 in [str(x) for x in part_lists[3]]:
                        self.append(IPv4AddressRange('.'.join([p1, p2, p3, p4])))

class IPv4AddressRange(object):
    """
    Defines a IPv4 address range, which you can:
    - check length of arbitrary range quickly
    - check if given address is in range
    - iterate to get IPv4Address objects for each address in range
    """

    def __init__(self, first, last=None):
        """
        First address and last address must be valid IPv4 addresses, and first
        address must be smaller than last address.

        If last is omitted and only first address is given, the value must be a
        valid subnet and the generated range covers all addresses, not just 'hosts':
        i.e. 192.168.0.0/24 == 192.168.0.0-192.168.0.255

        Any netmask given to the first or last address is ignored, i.e.
        IPv4AddressRange('192.168.0.0/24','192.168.0.10/8') returns range
        192.168.0.0-192.168.0.10

        Raises ValueError if the addresses can not be parsed or if the range is
        invalid.
        """
        self.__next = 0

        if isinstance(first, IPv4Address) and isinstance(last, IPv4Address):
            self.first = first
            self.last = last

        elif last is not None:
            self.first = IPv4Address(first)
            self.last = IPv4Address(last)

        elif first.count('/') == 1:
            # Support giving a subnet as range, but
            try:
                network = IPv4Address(first)
                self.first = IPv4Address(network.network)
                self.last = IPv4Address(network.broadcast)
            except ValueError:
                raise ValueError('Error parsing %s: %s' % (first,e))

        else:
            # Support nmap format like 1.2.3.1-254 for 1.2.3.1 to 1.2.3.254
            try:
                (start_ip,subnet_last) = first.split('-', 1)
                (start_ip, subnet_last) = first.split('-', 1)
                self.first = IPv4Address(start_ip)
                last = '%s.%s' % (
                    '.'.join(self.first.ipaddress.split('.')[:-len(subnet_last.split('.'))]),
                    subnet_last
                )
                self.last = IPv4Address(last.lstrip('.'))
            except ValueError,e:
                raise ValueError('Error parsing %s: %s' % (first,e))

        if self.last < self.first:
            raise ValueError('Invalid range: last address is smaller than first address')

    def __repr__(self):
        return '%s-%s' % (self.first.ipaddress, self.last.ipaddress)

    def __str__(self):
        return '%s-%s' % (self.first.ipaddress, self.last.ipaddress)

    def __len__(self):
        """
        Returns number of addresses in the range, including first and last address
        """
        return self.last.address - self.first.address + 1

    def __iter__(self):
        return self

    def next(self):
        if self.first.raw_value + self.__next > self.last.raw_value:
            raise StopIteration

        address = IPv4Address(self.first.raw_value + self.__next)
        self.__next += 1

        return address

    def contains(self, item):
        """
        Check if given address is in the range, including first and last
        address.

        The item must be IPv4Address object.
        """
        if item.address < self.first.address or item.address > self.last.address:
            return False
        return True

class IPv6Address(dict):
    def __init__(self, value):
        if value == 'default':
            value = '::0/0'

        try:
            address, bitmask = value.split('/')
        except ValueError:
            address = value
            bitmask = 128

        try:
            bitmask = int(bitmask)
            if int(bitmask) < 0 or int(bitmask) > 128:
                raise ValueError

        except ValueError:
            raise ValueError('Invalid IPv6 mask %s' % bitmask)

        try:
            subs = address.split(':')
            if len(subs) == 1:
                raise ValueError

            if subs[-1].count('.') == 3:
                # IPv6/IPv4 mapped address
                ipv4_mapping = IPv4Address(subs[-1])
                subs = subs[:-1] + [
                    ipv4_mapping.bitstring[2:6],
                    ipv4_mapping.bitstring[6:10],
                ]
            else:
                ipv4_mapping = None

            if subs.count('') == len(subs):
                # Address format like ::/96
                if int(bitmask) == 0:
                    bitstring = '0'
                else:
                    bitstring = '0' * (int(bitmask) / 4)

            elif subs.count('') > 0:
                # Address with shortcuts like (fe80::1/64)

                pad = subs.index('')
                if subs[pad+1] == '':
                    subs.pop(pad+1)

                if subs[0] != '':
                    start = ''.join(['%04x' % int(s, 16) for s in subs[:subs.index('')]])
                    end = ''.join(['%04x' % int(s, 16) for s in subs[subs.index('')+1:]])
                    bitstring = '%s%s%s' % (start, '0' * (32-len(start) - len(end)), end)
                else:
                    end = ''.join(['%04x' % int(s, 16) for s in subs[1:]])
                    bitstring = '%s%s' % ('0' * (32-len(end)), end)

            else:
                bitstring = ''.join(['%04x' % int(s, 16) for s in subs])[2:]

            hex_bitstring = '0x%s' % bitstring

            addrval = long(hex_bitstring, 16)
            network_bitstring = '0x%032x' % (
                addrval &~ ( U128_MAX & (2**(128-bitmask)-1) )
            )

            nbs = [int(network_bitstring[2:][i:i+4], 16) for i in range(0, 32, 4)]
            if len(nbs) > nbs.count(0):
                while nbs[-1] == 0:
                    nbs = nbs[:-1]

            network = '%s::/%s' % (':'.join(['%x' % n for n in nbs]), bitmask)

            try:
                revnibbles = '.'.join(
                    reversed([bitstring[x] for x in range(0, len(bitstring))])
                )
            except IndexError, emsg:
                raise ValueError('Error splitting bitstring for revnibbles: %s' % emsg)

        except ValueError,emsg:
            raise ValueError('Invalid IPv6 address: %s: %s' % (value, emsg))

        if not address.endswith('::'):
            self['type'] = 'address'
        else:
            self['type'] = 'subnet'

        if ipv4_mapping is not None:
            valid_translation_prefix = False
            for prefix in IPV6_IPV4_TRANSLATION_PREFIXES:
                if address[:len(prefix)] == prefix:
                    valid_translation_prefix = True
                    break
            if not valid_translation_prefix:
                raise ValueError('IPv6/IPv4 translation not supported for address %s' % address)

        self.update({
            'address':  address,
            'bitstring': hex_bitstring,
            'bitmask': bitmask,
            'network': network,
            'network_bitstring': network_bitstring,
            'revnibbles_int': '%s.ip6.int.' % revnibbles,
            'revnibbles_arpa': '%s.ip6.arpa.' % revnibbles,
        })


    @property
    def address(self):
        return self['address'].lower()

    @property
    def addressclass(self):
        for aclass, details in IPV6_ADDRESS_CLASS_MAP.items():
            for net in details['networks']:
                if IPv6Address(net).addressInNetwork(self.address):
                    return aclass

        return ADDRESS_CLASS_DEFAULT

    @property
    def bitstring(self):
        return self['bitstring']

    @property
    def bitmask(self):
        return self['bitmask']

    @property
    def network(self):
        return self['network']

    @property
    def network_bitstring(self):
        return self['network_bitstring']

    @property
    def revnibbles_arpa(self):
        return self['revnibbles_arpa']

    @property
    def revnibbles_int(self):
        return self['revnibbles_int']

    def __getattr__(self, attr):
        try:
            return self[attr]
        except KeyError:
            raise AttributeError('No such IPv6Address attribute: %s' % attr)

    def __hash__(self):
        return long(self.address)

    def __cmp__(self, other):
        if not hasattr(other, 'address'):
            return -1

        if self.address < other.address:
            return -1

        elif self.address > other.address:
            return 1

        return 0

    def __eq__(self,other):
        return self.__cmp__(other) == 0

    def __ne__(self,other):
        return self.__cmp__(other) != 0

    def __lt__(self, other):
        return self.__cmp__(other) < 0

    def __lte__(self, other):
        return self.__cmp__(other) <= 0

    def __gt__(self, other):
        return self.__cmp__(other) > 0

    def __gte__(self, other):
        return self.__cmp__(other) >= 0

    def __repr__(self):
        return '%s/%s' % (self.address, self.bitmask)

    def __addrfmt__(self, address, mask):
        s = '%032x' % address
        value = ['%x' % int(s[i:i+4],16) for i in range(0,32,4)]

        # Find the longest chain of 0's to truncate
        longest = 0
        index = None
        i = 0
        while i < len(value):
            if int(value[i],16) != 0:
                i+=1
                continue
            zeros = 0
            for v in value[i:]:
                if int(v,16) != 0:
                    break
                zeros += 1
            if zeros > longest:
                longest = zeros
                index = i
            i += zeros

        if index is not None:
            del(value[index:index+longest])
            value.insert(index,'')

        value = ':'.join(value)
        if value.startswith(':'):
            value = ':'+value
        if value.endswith(':'):
            value += ':'

        return '%s/%s' % (value,mask)

    @property
    def first(self):
        return IPv6Address(
            self.__addrfmt__(
                int(self.network_bitstring,16)+1,
                self.bitmask
            )
        )

    @property
    def last(self):
        return IPv6Address(
            self.__addrfmt__(
                int(self.network_bitstring,16)+2**(128-self.bitmask)-2,
                self.bitmask
            )
        )

    @property
    def broadcast(self):
        return IPv6Address(
            self.__addrfmt__(
                int(self.network_bitstring,16)+2**(128-self.bitmask)-1,
                self.bitmask
            )
        )

    @property
    def next(self):
        next = int(self.bitstring,16) + 1
        if next > U128_MAX:
            return None
        return IPv6Address(self.__addrfmt__(next,self.bitmask))

    @property
    def previous(self):
        next = int(self.bitstring,16) - 1
        if next < 0:
            return None
        return IPv6Address(self.__addrfmt__(next,self.bitmask))

    @property
    def next_network(self):
        network = int(self.network_bitstring,16) + 2**(128-self.bitmask)
        if network >= U128_MAX:
            return None
        return IPv6Address(self.__addrfmt__(network,self.bitmask))

    @property
    def previous_network(self):
        network = int(self.network_bitstring,16) - 2**(128-self.bitmask)
        if network < 0:
            return None
        return IPv6Address(self.__addrfmt__(network,self.bitmask))

    def addressInNetwork(self,value):
        """
        Tests if given IPv6 address is in range of this network,
        including network and broadcast addresses
        """
        if type(value) is not IPv6Address:
            try:
                value = IPv6Address(value)
            except ValueError,e:
                raise ValueError('Invalid IPv6Address: %s' % value)

        value = int(value.bitstring,16)
        first = int(self.network_bitstring,16)
        last = int(self.network_bitstring,16)+2**(128-self.bitmask)-1
        if value < first or value > last:
            return False

        return True

    def hostInNetwork(self, address):
        if type(address) is not IPv6Address:
            try:
                address = IPv6Address(address)
            except ValueError,e:
                raise ValueError('Invalid IPv6Address: %s' % address)

        address = int(address.bitstring, 16)
        first = int(self.network_bitstring, 16) + 1
        last = int(self.network_bitstring, 16) + 2**(128-self.bitmask) - 1
        if address < first or address > last:
            return False

        return True

class IPv4AddressList(dict):
    def __init__(self,values):
        for v in values:
            if not isinstance(v,IPv4Address):
                v = IPv4Address(v)

            if self.has_key(v.address):
                print 'Duplicate address'
                continue

            self[v.address] = v

    def keys(self):
        return sorted(map(lambda a: int(a), dict.keys(self)))

    def values(self):
        return [self[a] for a in map(lambda a: int(a), self.keys())]

    def items(self):
        return [(a,self[a]) for a in map(lambda a: int(a), self.keys())]

    def sorted_shortened(self):
        """
        Return the entries in IPv4Address, sorted by address and compressed
        to IPv4Range objects when consequetive addresses are found.

        Returns mixed list of IPv4Address and IPv4Range objects.
        """
        addresses = []
        iprange = []
        for v in self.values():
            if iprange == []:
                iprange = [v]
                continue

            if v.address == iprange[-1].address+1:
                iprange.append(v)
                continue

            if len(iprange) > 1:
                addresses.append(IPv4AddressRange(
                    iprange[0].ipaddress,
                    iprange[-1].ipaddress,
                ))
            elif len(iprange) == 1:
                addresses.append(iprange[0])
            else:
                raise NotImplementedError('IMPOSSIBLE BUG: %s' % iprange)

            iprange = [v]

        if len(iprange) > 1:
            addresses.append(IPv4AddressRange(iprange[0].ipaddress, iprange[-1].ipaddress))

        elif len(iprange) == 1:
            addresses.append(iprange[0])

        return addresses

class SubnetPrefixIterator(object):
    def __init__(self, address, splitmask):
        try:
            splitmask = int(splitmask)
        except ValueError:
            raise ValueError('Invalid splitmask')

        try:
            self.address = IPv4Address(address)
            self.last = self.address.bitmask <= 29 and self.address.last.address or None
        except ValueError,emsg:
            try:
                self.address = IPv6Address(address)
                self.last = long(self.address.last.bitstring, 16)
            except ValueError:
                raise ValueError('Not valid IPv4 or IPv6 address: %s' % address)

        if isinstance(self.address,IPv4Address) and self.address.bitmask>=30:
            raise ValueError("Can't split address with mask %s" % self.address.bitmask)
        if self.address.bitmask >= splitmask:
            raise ValueError('Split mask must be smaller than network mask')

        if isinstance(self.address, IPv4Address):
            self.first = IPv4Address('%s/%s' % (self.address.network, splitmask))
        if isinstance(self.address,IPv6Address):
            a = self.address.network.split('/')[0]
            self.first = IPv6Address('%s/%s' % (a, splitmask))
        self.__next = self.first

    def __iter__(self):
        return self

    def next(self):
        try:
            if type(self.__next) == IPv4Address:
                if self.__next is None:
                    raise StopIteration
                entry = self.__next
                if self.address.last.raw_value <= entry.first.raw_value:
                    raise StopIteration
                self.__next = entry.next_network

            if type(self.__next) == IPv6Address:
                if self.__next is None:
                    raise StopIteration
                entry = self.__next
                entry_first = long(entry.first.bitstring, 16)
                if self.last <= entry_first:
                    raise StopIteration
                self.__next = entry.next_network

        except StopIteration:
            self.__next = self.first
            raise StopIteration

        return entry

def parse_address(value):
    """Parse address

    Parse address to IPv4Address, IPv6Address or EthernetMACAddress

    """
    for fmt in (IPv4Address, IPv6Address, EthernetMACAddress):
        try:
            return fmt(value)
        except ValueError:
            pass

    raise ValueError('Unknown address format: %s' % value)



"""
DNS Zone authorized server processing classes
"""

from seine.dns import DNSError
from seine.dns.resolver import resolve_records, QueryError
from seine.dns.delegation import DNSZoneDelegation

class AuthorizedZoneServers(object):
    """DNS zone authorized servers

    Parser for authorized servers for DNS domain

    """
    def __init__(self, domain, rootfile='/tmp/db.root', timeout=5, ipv4=True, ipv6=False):
        self.domain = domain
        self.timeout = timeout

        self.SOA = {}
        self.NS  = {}

        if ipv4:
            self.ipv4_delegation = DNSZoneDelegation(
                domain, rrtype='A', rootfile=rootfile, timeout=timeout
            )
        else:
            self.ipv4_delegation = None

        if ipv6:
            self.ipv6_delegation = DNSZoneDelegation(
                domain, rrtype='AAAA', rootfile=rootfile, timeout=timeout
            )
        else:
            self.ipv6_delegation = None

    def update_delegations(self):
        """Update zone delegations

        Query and update delegations for zone

        """
        if self.ipv4_delegation is not None:
            self.ipv4_delegation.update_delegations()

        if self.ipv6_delegation is not None:
            self.ipv6_delegation.update_delegations()

    def query_server(self, server, rrtype):
        """Query RR for server

        Resolve this zone's RR from given server

        """
        return resolve_records(
            query=self.domain,
            rrtype=rrtype,
            nameserver=server,
            timeout=self.timeout
        )

    def validate_SOA(self):
        """Validate domain SOA

        Validate SOA record of a domain: this is only example stub right now!

        TODO - actually compare SOA records with each other

        """
        if not self.ipv4_delegation and not self.ipv6_delegation:
            self.update_delegations()

        soa_records = {}
        if self.ipv4_delegation is not None:
            for delegation in self.ipv4_delegation:
                if delegation.name not in soa_records.keys():
                    soa_records[delegation.name] = delegation.soa

        if self.ipv6_delegation is not None:
            for server in self.ipv6_delegation:
                if delegation.name not in soa_records.keys():
                    soa_records[delegation.name] = delegation.soa

        for server, server_soa in soa_records.items():
            for other, soa in soa_records.items():
                if server == other:
                    continue

                if server_soa != soa:
                    raise DNSError('SOA records do not match: %s and %s' % (server, delegation.soa))

    def validate_NS(self):
        """Validate zone NS records

        Validate NS records of a domain: this is only example stub right now!

        TODO - actually compare NS records with each other and TLD delegation

        """
        if self.ipv4_delegation:
            for delegation in self.ipv4_delegation:
                if self.ipv4_delegation.delegated_servers != delegation.nameservers:
                    raise DNSError('%s: servers from delegation and server differ: %s and %s' % (
                        delegation,
                        ','.join('%s' % s for s in self.ipv4_delegation.delegated_servers),
                        ','.join('%s' % s for s in delegation.nameservers)
                    ))

        if self.ipv6_delegation:
            for delegation in self.ipv6_delegation:
                if self.ipv6_delegation.delegated_servers != delegation.nameservers:
                    raise DNSError('%s: servers from delegation and server differ: %s and %s' % (
                        delegation,
                        ','.join('%s' % s for s in self.ipv6_delegation.delegated_servers),
                        ','.join('%s' % s for s in delegation.nameservers)
                    ))



"""
Wrappers related to net-snmp python tools, making it's usage much easier.
"""

__all__ = ['agent', 'client', 'script']

# Supported SNMP versions
SNMP_VERSIONS = ( '1', '2c', '3' )

class SNMPError(Exception):
    pass

def cmp_oid(v1, v2):
    """
    Compare function to compare two oids given as strings or lists of integers
    """
    try:
        if type(v1) == list:
            v1 = [int(i) for i in v1]
        else:
            v1 = [int(i) for i in v1.lstrip('.').split('.')]

        if type(v2) == list:
            v2 = [int(i) for i in v2]
        else:
            v2 = [int(i) for i in v2.lstrip('.').split('.')]

    except ValueError:
        raise SNMPError('Unsupported OID values to cmp_oid')

    for i in range( min(len(v1), len(v2)) ):
        r = cmp(v1[i], v2[i])
        if r != 0:
            return r

    if len(v1) < len(v2):
        return -1
    elif len(v2) > len(v1):
        return 1

    return 0




"""
SNMP agent base class for net-snmp pass/pass_persist agents.

See the SNMPAgentTest class below how to use the SNMPAgent class.
"""

import sys
import os
import re
import time
import signal
import errno
import select

from systematic.shell import Script, ScriptThread, ScriptError
from systematic.log import Logger, LoggerError
from seine.address import IPv4Address
from seine.snmp import SNMPError

# Valid SNMP data types
SNMP_DATA_TYPE_MAP = {
    'integer':      lambda x: int(x),
    'gauge':        lambda x: int(x),
    'counter':      lambda x: int(x),
    'timeticks':    lambda x: int(x),
    'string':       lambda x: str(x),
    'ipaddress':    lambda x: str(IPv4Address(x)),
    'objectid':     lambda x: str(x),
}


class OIDPrefix(object):
    """Sortable OID base class

    Base class for OID prefixes, sortable in OID order

    """
    def __init__(self, oid):
        self.log = Logger('snmp').default_stream
        self.oid = self.__format_oid__(oid)
        self.oid_string = self.__format_oid_string__(oid)

        self._parent = None
        self._next = None

    def __format_oid_string__(self, oid):
        return '.%s' % '.'.join(str(x) for x in self.__format_oid__(oid))

    def __format_oid__(self, oid):
        """Format OID

        Format OID string to validated OID list

        Return list of integeres
        """

        if isinstance(oid, basestring):
            try:
                oid = oid.strip('.').split('.')
            except ValueError:
                raise ValueError('Invalid OID: %s' % oid)

        return self.__validate_oid__(oid)

    def __validate_oid__(self, oid):
        """Validate OID and return

        Validate list of integers to be a valid OID

        Returns value if OK
        """
        try:
            oid = [int(x) for x in oid]
            for x in oid:
                if x<=0:
                    raise ValueError

        except ValueError:
            raise ValueError('Invalid OID: %s' % oid)

        return oid

    def __invalid__(self, message):
        """Invalid request

        Log error and return None

        """
        self.log.debug('%s' % message)
        return None

    def __cmp__(self, other):
        """Compare oid prefix items

        Compares OID prefixes

        """
        if isinstance(other, basestring):
            oid = self.__format_oid__(other)

        elif isinstance(other, OIDPrefix):
            oid = other.oid

        else:
            raise ValueError('Error comparing OIDPrefix to %d' % type(other))

        if len(oid) != len(self.oid):
            return cmp(len(self.oid), len(oid))

        for i in range(0, len(oid)):
            if oid[i] != self.oid[i]:
                return cmp(self.oid[i], oid[i])

        return cmp(self.oid, oid)

    def __eq__(self, oid):
        return self.__cmp__(oid) == 0

    def __ne__(self, oid):
        return self.__cmp__(oid) != 0

    def __lt__(self, oid):
        return self.__cmp__(oid) < 0

    def __le__(self, oid):
        return self.__cmp__(oid) <= 0

    def __gt__(self, oid):
        return self.__cmp__(oid) > 0

    def __ge__(self, oid):
        return self.__cmp__(oid) >= 0

    @property
    def get_response(self):
        return None

    @property
    def next_response(self):
        return None

    @property
    def parent(self):
        return self._parent
    @parent.setter
    def parent(self, value):
        self._parent = value

    @property
    def next(self):
        return self._next
    @next.setter
    def next(self, value):
        self._next = value

    def match(self, other):
        """Match other OID

        Match OID prefixes. Default function checks if self.oid == other.oid

        Override in child class (for example, trees match tree prefix)
        """
        return self == other


class Item(OIDPrefix):
    def __init__(self, oid, key, value=None, readonly=True):
        OIDPrefix.__init__(self, oid)
        self.readonly = readonly

        if key in SNMP_DATA_TYPE_MAP.keys():
            self.key = key
        else:
            raise SNMPError('Invalid item type: %s' % key)

        if value is not None:
            try:
                self.value = SNMP_DATA_TYPE_MAP[self.key](value)
            except:
                raise SNMPError('Invalid %s item value "%s"' % (self.key, value))
        else:
            self.value = None

    def __repr__(self):
        return '%s %s %s' % (self.oid_string, self.key, self.value)

    @property
    def get_response(self):
        """GET response for an item

        Strings required for formatted GET response

        """
        return '%s\n%s\n%s' % (self.oid_string, self.key, self.value)

    @property
    def next_response(self):
        """NEXT response for an item

        Strings required for formatted NEXT response

        """
        return '%s\n%s\n%s' % (self.oid_string, self.key, self.value)

    def check_length(self, value):
        """Check SET value length

        Override in child class to check SET value length

        """
        return True

    def check_value(self, value):
        """Check SET value

        Override in child class to check SET value

        """
        return True

    def check_value_consistency(self, value):
        """Check SET value consistency

        Override in child class to check SET value consistency

        """
        return True

class Tree(OIDPrefix):
    """SNMP agent OID Tree

    Tree of SNMP OIDs. Can contain items or subtrees

    """
    def __init__(self, oid):
        OIDPrefix.__init__(self, oid)

        self.items = []
        self.item_index = {}

        self.subtrees = []

    def __repr__(self):
        return self.oid_string

    def __register__(self, entry, index=0, next=None):
        """Register wrapper

        Method to actually register things:
        - register trees if given
        - add OID to item_index
        - link to previous item's next

        """
        entry.parent = self

        if index > 0:
            previous = self.items[index - 1]
            previous.next = entry
        else:
            previous = None
        entry.next = next

        self.items.append(entry)
        self.item_index[entry.oid_string] = entry

        if isinstance(entry, Tree):
            self.subtrees.append(entry)

        self.items.sort()
        return entry

    def clear(self):
        """Clear items

        Clear items from tree, preserve subtrees

        """
        for oid in sorted(self.item_index.keys()):
            item = self.item_index[oid]
            if isinstance(item, Item):
                del self.item_index[oid]
            elif isinstance(item, Tree):
                item.clear()

        for item in [x for x in self.items]:
            if isinstance(item, Item):
                self.items.remove(item)
            elif isinstance(item, Tree):
                item.clear()

    def relative_oid(self, oid):
        if oid == self.oid:
            raise SNMPError('Attempt to get relative oid for tree root: %s' % self.oid)

        if oid[:len(self.oid)] != self.oid:
            raise SNMPError('OID out of tree %s: %s' % (self.oidstring, oid))

        return oid[len(self.oid):]

    def match(self, other):
        """Check OID prefix match

        Check if given object matches our OID prefix
        """
        if isinstance(other, basestring):
            oid = self.__format_oid__(other)

        elif isinstance(other, list):
            oid = self.__validate_oid__(other)

        elif isinstance(other, OIDPrefix):
            oid = other.oid

        else:
            raise SNMPError('UNSUPPORTED MATCH TYPE: %s' % type(other))

        if len(oid) < len(self.oid):
            return False

        if oid[:len(self.oid)] != self.oid:
            return False

        return True

    def register_tree(self, oid):
        """Register subtree

        Register a new subtree with given OID

        Returns the created Tree object
        """
        return self.add(Tree(oid))

    def register(self, oid, key, value, readonly=True):
        """Register Item to tree

        Register new Item to the tree with provided oid, key, value

        Returns the created item
        """
        return self.add(Item(oid, key, value))

    def add(self, item):
        """Add an item to tree

        Add an item to the tree. Item must be OIDPrefix instance.

        Returns the registered item.
        """
        if not isinstance(item, OIDPrefix):
            raise SNMPError('Can only add OIDPrefix objects to tree')

        if not self.match(item):
            raise SNMPError('OID prefix does not match: %s' % item.oid_string)

        for i, existing in enumerate(self.items):

            if isinstance(existing, Tree) and existing.match(item):
                return existing.add(item)

            if item < existing:
                return self.__register__(item, i, existing)

            if item == existing:
                raise SNMPError('OID already registered: %s' % item.oid)

        return self.__register__(item, len(self.items))

    def add_values(self, key, values, index=None, readonly=True):
        """Add a list of values

        Add a list of values as Item entries with given key, starting
        from given index. If index is None, first index is last + 1

        """
        if key not in SNMP_DATA_TYPE_MAP:
            raise SNMPError('Invalid key %s' % key)

        if not isinstance(values, list):
            raise SNMPError('Values must be a list of simple values: %s' % values)

        if index is None:
            if self.items:
                self.log.debug('Lookup relative OID from %s' % self.items[-1])
                index = self.relative_oid(self.items[-1].oid)
                index[-1] += 1
            else:
                index = [1]

        for value in values:
            oid = [x for x in self.oid] + index
            self.register(oid, key, value, readonly)
            index[-1] += 1

        return self

    def add_prefix_map(self, prefixes):
        """Add a list of mapped prefixes

        Add a list of mapped prefixes, for example:

            tree = agent.register_tree('1.2.3.4.4')
            tree.add_prefix_map([
                { 'oid': '1.2.3.4.4.1', 'key': 'string', 'values': [x for x in string.letters[:10]]},
                { 'oid': '1.2.3.4.4.2', 'key': 'integer', 'readonly': False, 'values': [x for x in string.digits[:10]]},
            ])

            Each list must have equal number of values: think of ifIndex and ifInOctets

            Returns nothing
        """
        try:
            if not isinstance(prefixes, list):
                raise ValueError('Prefixes must be a list')

            # Validate prefixes before insert
            length = None
            for prefix in prefixes:
                if not isinstance(prefix, dict):
                    raise ValueError

                if 'oid' in prefix:
                    prefix['oid'] = self.__format_oid__(prefix['oid'])
                    self.relative_oid(prefix['oid'])
                else:
                    raise ValueError('Missing oid: %s' % prefix)

                if 'key' in prefix:
                    try:
                        callback = SNMP_DATA_TYPE_MAP[prefix['key']]
                    except KeyError:
                        raise ValueError('Invalid prefix key %s' % prefix)
                else:
                    raise ValueError('Missing key: %s' % prefix)

                if 'readonly' in prefix:
                    prefix['readonly'] = prefix['readonly'] and True or False
                else:
                    prefix['readonly'] = True

                if 'values' in prefix:
                    items = prefix['values']

                    if not isinstance(items, list):
                        raise ValueError('Items not a list')

                    if len(items)==0 or length is not None and len(items) != length:
                        raise ValueError('Invalid values list length: %s' % prefix)

                else:
                    raise ValueError('Missing values: %s' % prefix)

        except ValueError:
            raise SNMPError('Invalid prefixes format: %s' % prefixes)

        for prefix in prefixes:
            tree = self.add(Tree(prefix['oid']))
            tree.add_values(prefix['key'], prefix['values'], readonly=prefix['readonly'])

    def GET(self, oid):
        """Return OID from tree

        Return value for matching OID from tree or None

        """
        try:
            formatted = self.__format_oid__(oid)
            oidstring = self.__format_oid_string__(oid)
        except ValueError:
            return self.__invalid__('GET invalid OID %s' % oid)

        if formatted == self.oid:
            return self.__invalid__('GET for OID tree root %s' % oidstring)

        try:
            entry = self.item_index[oidstring]

            if isinstance(entry, Tree):
                return entry.GET(oid)

            if isinstance(entry, Item):
                if entry.value is not None:
                    return entry
                else:
                    return __invalid__('Tree %s entry %s not initialized' % (self, entry))

        except KeyError:
            pass

        for tree in self.subtrees:
            if tree.match(formatted):
                return tree.GET(oid)

        return None

    def NEXT(self, oid):
        """Return next OID from tree

        Return value for matching next OID from tree or None

        """
        try:
            formatted = self.__format_oid__(oid)
        except ValueError:
            return self.__invalid__('GET invalid OID')

        if formatted == self.oid:
            if not len(self.items):
                return self.__invalid__('NEXT for root of empty tree %s' % self.oidstring)
            next = self.items[0]

            if isinstance(next, Tree):
                return next.NEXT(next.oid_string)

            if hasattr(next, 'value'):
                return next

            else:
                raise SNMPError('ERROR getting NEXT from %s' % next)

        for tree in self.subtrees:
            if tree.match(oid):
                return tree.NEXT(oid)

        try:
            index = self.__format_oid_string__(oid)
        except ValueError:
            return self.__invalid__('NEXT invalid OID %s' % oid)

        try:
            entry = self.item_index[index]
        except KeyError:
            if self.items:
                return self.items[0]
            else:
                return self.__invalid__('NEXT unknown OID %s' % oid)

        if isinstance(entry, Tree):
            return entry.NEXT(oid)

        if isinstance(entry.next, Tree):
            return entry.next.NEXT(oid)

        if entry.next:
            return entry.next
        elif self.next:
            return self.next.NEXT(oid)
        else:
            self.__invalid__('NEXT unknown OID %s' % oid)

    def SET(self, oid, value):
        """Set OID value

        Set value for OID in tree. Item matching OID must exist before SET.

        Returns error code string or DONE if successful
        """
        entry = self.GET(oid)
        if entry is None:
            return 'not-writable'

        if not hasattr(entry, 'value') or entry.readonly:
            self.log.debug('error writing entry %s: %s %s' % (entry, type(entry), hasattr(entry, 'value')))
            return 'not-writable'

        try:
            value = SNMP_DATA_TYPE_MAP[entry.key](value)
        except ValueError:
            return 'wrong-type'

        if not entry.check_length(value):
            return 'wrong-length'

        if not entry.check_value(value):
            return 'wrong-value'

        if not entry.check_value_consistency(value):
            return 'inconsistent-value'

        entry.value = value
        return 'DONE'

class SNMPAgent(Script):
    """SNMP Agent implementation

    SNMP agent daemon for net-snmpd pass_persist mode

    Example usage:

        agent = SNMPAgent('.1.2.3.4')
        tree = agent.register_tree('1.2.3.4.1')
        tree.register('1.2.3.4.1.1', 'integer', 1)
        tree.register('1.2.3.4.1.2', 'integer', 2)

        agent.register_tree('1.2.3.4.2')
        agent.register('1.2.3.4.2.1', 'integer', 1)
        agent.register('1.2.3.4.2.4', 'integer', 4)
        agent.register('1.2.3.4.2.3', 'integer', 3)
        agent.register('1.2.3.4.2.2', 'integer', 2)

        tree = agent.register_tree('1.2.3.4.3')
        tree.register('1.2.3.4.3.1.1', 'integer', 1)
        tree.add_values('string', [x for x in string.digits[2:]])

        tree = agent.register_tree('1.2.3.4.4')
        tree.add_prefix_map([
            { 'oid': '1.2.3.4.4.1', 'key': 'string', 'values': [x for x in string.letters[:10]]},
            { 'oid': '1.2.3.4.4.2', 'key': 'integer', 'values': [x for x in string.digits[:10]]},
        ])

        agent.run()
    """
    def __init__(self, oid, reload_interval=60):
        Script.__init__(self)
        self.log = Logger('snmp').default_stream

        self.reload_interval = reload_interval
        self.last_reload = None

        self.add_argument('-g', '--get', help='SNMP GET request')
        self.add_argument('-n', '--next', help='SNMP GET request')
        self.add_argument('-t', '--tree', action='store_true', help='Show OID tree')

        self.tree = Tree(oid)

    def __SIGHUP__(self, signum, frame):
        """
        Signal handler to reload configuration. Note this requires also the
        IOError processing in main input loop below
        """
        self.log.debug('Reloading from signal')
        self.reload()

    def register_tree(self, oid):
        return self.tree.add(Tree(oid))

    def register(self, oid, key, value):
        return self.tree.add(Item(oid, key, value))

    def GET(self, oid):
        """GET from agent

        Return given OID from agent tree or None

        """
        try:
            return self.tree.GET(oid)
        except SNMPError, emsg:
            self.log.debug(emsg)
            return None

    def NEXT(self, oid):
        """NEXT from agent

        Return given NEXT from agent tree or None

        """
        try:
            return self.tree.NEXT(oid)
        except SNMPError, emsg:
            self.log.debug(emsg)
            return None

    def SET(self, oid, value):
        """SET registered OID

        Set value for registered OID in agent tree.

        Note: By default items registered are readonly.


        """
        return self.tree.SET(oid, value)

    def clear(self):
        return self.tree.clear()

    def reload(self):
        """
        This method must be implemented in a child class. It is used to
        reload the SNMP tree data from files, if possible.
        """
        raise NotImplementedError('You must implement reload in child class')

    def run(self):
        """SNMP Agent main loop

        Main loop to execute for agent. You can either run this in:

        - 'pass' mode: get/next for single OID by passing in OptionParser
          'options' value 'get' or 'next' (snmpd.conf 'pass' agent)
        - 'pass_persist' mode: without any options, in which case the loop
           acts as permanent snmpd pass_persist agent.
        """

        signal.signal(signal.SIGHUP, self.__SIGHUP__)
        args = self.parse_args()

        if args.tree:
            self.message(self.tree.oid_string)

        elif args.get:
            entry = self.GET(args.get)
            if entry is not None:
                sys.stdout.write('%s\n' % entry.get_response)

        elif args.next:
            entry = self.NEXT(args.next)
            if entry is not None:
                sys.stdout.write('%s\n' % entry.next_response)

        if args.tree or args.get or args.next:
            self.exit(0)

        EOF = ''
        self.log.debug('Starting SNMP agent for OID %s' % self.tree.oid)

        while True:
            ready = select.select([sys.stdin], [], [], 0.2)[0]
            if not ready:
                if self.last_reload is not None:
                    since_reload = time.time() - self.last_reload
                    if since_reload > self.reload_interval:
                        self.reload()
                        self.last_reload = time.time()
                else:
                    self.last_reload = time.time()
                continue

            try:
                cmd = sys.stdin.readline()
                if cmd == EOF:
                    break
                if cmd.strip() == '':
                    self.log.debug('Received agent shutdown')
                    break

                cmd = cmd.rstrip().lower()

                if cmd == 'quit':
                    return

                if cmd == 'reload':
                    self.reload()

                if cmd == 'ping':
                    sys.stdout.write('PONG\n')

                if cmd in ('set', 'get', 'getnext'):
                    oid = sys.stdin.readline()
                    if oid == EOF:
                        break
                    oid = oid.rstrip()

                if cmd == 'set':
                    value = sys.stdin.readline()
                    if value == EOF:
                        break
                    result = self.SET(oid, value)
                    sys.stdout.write('%s\n' % result)

                elif cmd == 'get':
                    entry = self.GET(oid)
                    if entry is not None:
                        sys.stdout.write('%s\n' % entry.get_response)
                    else:
                        sys.stdout.write('NONE\n')

                elif cmd == 'getnext':
                    entry = self.NEXT(oid)
                    if entry is not None:
                        sys.stdout.write('%s\n' % entry.next_response)
                    else:
                        sys.stdout.write('NONE\n')

                else:
                    self.log.debug('Unknown command (should be get/set/getnext)')
                sys.stdout.flush()

            except IOError, emsg:
                # we get EINTR with SIGHUP and we can ignore it
                if emsg[0]==errno.EINTR:
                    continue
                self.log.debug('IOError: %s' % emsg[1])
                return

            except KeyboardInterrupt:
                # Interactive mode, user interrupted
                self.log.debug('Quitting...')
                return
